/* Fo
 * fo-property-border.c: border property
 *
 * Copyright (C) 2001 Sun Microsystems
 *
 * $Id: fo-property-border.c,v 1.5 2004/03/23 11:59:23 tonygraham Exp $
 *
 * See Copyright for the status of this software.
 */

#include <string.h>
#include "libfo/fo-utils.h"
#include <fo-datatype.h>
#include <fo-property.h>
#include <fo-property-border-private.h>
#include <fo-property-border-before-color.h>
#include <fo-property-border-before-style.h>
#include <fo-property-border-before-width.h>
#include <fo-property-border-after-color.h>
#include <fo-property-border-after-style.h>
#include <fo-property-border-after-width.h>
#include <fo-property-border-start-color.h>
#include <fo-property-border-start-style.h>
#include <fo-property-border-start-width.h>
#include <fo-property-border-end-color.h>
#include <fo-property-border-end-style.h>
#include <fo-property-border-end-width.h>
#include <fo-property-border-top-color.h>
#include <fo-property-border-top-style.h>
#include <fo-property-border-top-width.h>
#include <fo-property-border-bottom-color.h>
#include <fo-property-border-bottom-style.h>
#include <fo-property-border-bottom-width.h>
#include <fo-property-border-left-color.h>
#include <fo-property-border-left-style.h>
#include <fo-property-border-left-width.h>
#include <fo-property-border-right-color.h>
#include <fo-property-border-right-style.h>
#include <fo-property-border-right-width.h>
#include <fo-property-color.h>

#ifndef FO_PROPERTY_BORDER_WIDTH_THIN
#define FO_PROPERTY_BORDER_WIDTH_THIN	0.5
#endif

#ifndef FO_PROPERTY_BORDER_WIDTH_MEDIUM
#define FO_PROPERTY_BORDER_WIDTH_MEDIUM	2.0
#endif

#ifndef FO_PROPERTY_BORDER_WIDTH_THICK
#define FO_PROPERTY_BORDER_WIDTH_THICK	4.0
#endif

/* Inherited: FALSE */
/* Shorthand: FALSE */
/* border-width */
/* <border-width> | inherit */
GType
fo_enum_border_width_get_type (void)
{
  static GType etype = 0;
  if (etype == 0)
    {
      static const GEnumValue values[] = {
        { FO_ENUM_BORDER_WIDTH_THIN, "FO_ENUM_BORDER_WIDTH_THIN", "thin" },
        { FO_ENUM_BORDER_WIDTH_MEDIUM, "FO_ENUM_BORDER_WIDTH_MEDIUM", "medium" },
        { FO_ENUM_BORDER_WIDTH_THICK, "FO_ENUM_BORDER_WIDTH_THICK", "thick" },
        { 0, NULL, NULL }
      };
      etype = g_enum_register_static ("FoEnumBorderWidth", values);
    }
  return etype;
}

/* border-style */
/* <border-style> | inherit */
GType
fo_enum_border_style_get_type (void)
{
  static GType etype = 0;
  if (etype == 0)
    {
      static const GEnumValue values[] = {
        { FO_ENUM_BORDER_STYLE_NONE, "FO_ENUM_BORDER_STYLE_NONE", "none" },
        { FO_ENUM_BORDER_STYLE_HIDDEN, "FO_ENUM_BORDER_STYLE_HIDDEN", "hidden" },
        { FO_ENUM_BORDER_STYLE_DOTTED, "FO_ENUM_BORDER_STYLE_DOTTED", "dotted" },
        { FO_ENUM_BORDER_STYLE_DASHED, "FO_ENUM_BORDER_STYLE_DASHED", "dashed" },
        { FO_ENUM_BORDER_STYLE_SOLID, "FO_ENUM_BORDER_STYLE_SOLID", "solid" },
        { FO_ENUM_BORDER_STYLE_DOUBLE, "FO_ENUM_BORDER_STYLE_DOUBLE", "double" },
        { FO_ENUM_BORDER_STYLE_GROOVE, "FO_ENUM_BORDER_STYLE_GROOVE", "groove" },
        { FO_ENUM_BORDER_STYLE_RIDGE, "FO_ENUM_BORDER_STYLE_RIDGE", "ridge" },
        { FO_ENUM_BORDER_STYLE_INSET, "FO_ENUM_BORDER_STYLE_INSET", "inset" },
        { FO_ENUM_BORDER_STYLE_OUTSET, "FO_ENUM_BORDER_STYLE_OUTSET", "outset" },
        { 0, NULL, NULL }
      };
      etype = g_enum_register_static ("FoEnumBorderStyle", values);
    }
  return etype;
}

static void fo_property_border_init         (FoPropertyBorder      *property_border);
static void fo_property_border_class_init   (FoPropertyBorderClass *klass);
static void fo_property_border_base_init   (FoPropertyBorderClass *klass);
static void fo_property_border_base_finalize (FoPropertyBorderClass *klass);
static void fo_property_border_finalize     (GObject       *object);

static gpointer parent_class;

static FoDatatype *width_thin;
static FoDatatype *width_medium;
static FoDatatype *width_thick;
static FoDatatype *style_none;
static FoDatatype *style_hidden;
static FoDatatype *style_dotted;
static FoDatatype *style_dashed;
static FoDatatype *style_solid;
static FoDatatype *style_double;
static FoDatatype *style_groove;
static FoDatatype *style_ridge;
static FoDatatype *style_inset;
static FoDatatype *style_outset;

/**
 * fo_property_border_get_type:
 * 
 * Register the #FoPropertyBorder type if not already registered and
 * return its #GType value.
 * 
 * Return value: #GType of #FoPropertyBorder.
 **/
GType
fo_property_border_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
      {
        sizeof (FoPropertyBorderClass),
        (GBaseInitFunc) fo_property_border_base_init,
        (GBaseFinalizeFunc) fo_property_border_base_finalize,
        (GClassInitFunc) fo_property_border_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data */
        sizeof (FoPropertyBorder),
        0,              /* n_preallocs */
        (GInstanceInitFunc) fo_property_border_init,
	NULL		/* value_table */
      };
      
      object_type = g_type_register_static (FO_TYPE_PROPERTY,
                                            "FoPropertyBorder",
                                            &object_info, 0);
    }
  
  return object_type;
}

/**
 * fo_property_border_init:
 * @border: #FoPropertyBorder object to initialise.
 * 
 * Implements #GInstanceInitFunc for #FoPropertyBorder.
 **/
void
fo_property_border_init (FoPropertyBorder *border)
{
  FO_PROPERTY (border)->value =
    NULL;
}

/**
 * fo_property_border_base_init:
 * @klass: #FoPropertyBorderClass object to initialise.
 * 
 * Implements #GBaseInitFunc for #FoPropertyBorderClass.
 **/
void
fo_property_border_base_init (FoPropertyBorderClass *klass)
{
  width_thin =
    g_object_new (FO_TYPE_LENGTH,
		  "value",
		  FO_PROPERTY_BORDER_WIDTH_THIN,
		  NULL);
  width_medium =
    g_object_new (FO_TYPE_LENGTH,
		  "value",
		  FO_PROPERTY_BORDER_WIDTH_MEDIUM,
		  NULL);
  width_thick =
    g_object_new (FO_TYPE_LENGTH,
		  "value",
		  FO_PROPERTY_BORDER_WIDTH_THICK,
		  NULL);
  style_none =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_BORDER_STYLE),
                  "value",
                  FO_ENUM_BORDER_STYLE_NONE,
                  NULL);
  style_hidden =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_BORDER_STYLE),
                  "value",
                  FO_ENUM_BORDER_STYLE_HIDDEN,
                  NULL);
  style_dotted =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_BORDER_STYLE),
                  "value",
                  FO_ENUM_BORDER_STYLE_DOTTED,
                  NULL);
  style_dashed =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_BORDER_STYLE),
                  "value",
                  FO_ENUM_BORDER_STYLE_DASHED,
                  NULL);
  style_solid =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_BORDER_STYLE),
                  "value",
                  FO_ENUM_BORDER_STYLE_SOLID,
                  NULL);
  style_double =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_BORDER_STYLE),
                  "value",
                  FO_ENUM_BORDER_STYLE_DOUBLE,
                  NULL);
  style_groove =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_BORDER_STYLE),
                  "value",
                  FO_ENUM_BORDER_STYLE_GROOVE,
                  NULL);
  style_ridge =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_BORDER_STYLE),
                  "value",
                  FO_ENUM_BORDER_STYLE_RIDGE,
                  NULL);
  style_inset =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_BORDER_STYLE),
                  "value",
                  FO_ENUM_BORDER_STYLE_INSET,
                  NULL);
  style_outset =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_BORDER_STYLE),
                  "value",
                  FO_ENUM_BORDER_STYLE_OUTSET,
                  NULL);
}

/**
 * fo_property_border_base_finalize:
 * @klass: #FoPropertyBorderClass object to finalise.
 * 
 * Implements #GBaseFinalizeFunc for #FoPropertyBorderClass.
 **/
void
fo_property_border_base_finalize (FoPropertyBorderClass *klass)
{
  g_object_unref (width_thin);
  g_object_unref (width_medium);
  g_object_unref (width_thick);

  g_object_unref (style_none);
  g_object_unref (style_hidden);
  g_object_unref (style_dotted);
  g_object_unref (style_dashed);
  g_object_unref (style_solid);
  g_object_unref (style_double);
  g_object_unref (style_groove);
  g_object_unref (style_ridge);
  g_object_unref (style_inset);
  g_object_unref (style_outset);
}

/**
 * fo_property_border_class_init:
 * @klass: #FoPropertyBorderClass object to initialise.
 * 
 * Implements #GClassInitFunc for #FoPropertyBorderClass.
 **/
void
fo_property_border_class_init (FoPropertyBorderClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  
  parent_class = g_type_class_peek_parent (klass);
  
  object_class->finalize = fo_property_border_finalize;
}

/**
 * fo_property_border_finalize:
 * @object: #FoPropertyBorder object to finalize.
 * 
 * Implements #GObjectFinalizeFunc for #FoPropertyBorder.
 **/
void
fo_property_border_finalize (GObject *object)
{
  FoPropertyBorder *border;

  border = FO_PROPERTY_BORDER (object);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * fo_property_border_new:
 * 
 * Creates a new #FoPropertyBorder initialized to default value.
 * 
 * Return value: the new #FoPropertyBorder.
 **/
FoProperty*
fo_property_border_new (void)
{
  FoProperty* border;

  border =
    FO_PROPERTY (g_object_new (fo_property_border_get_type (),
			       NULL));
  
  return border;
}

/**
 * fo_property_border_resolve:
 * @current_context: #FoContext of the current #FoFo.
 * @parent_context:  #FoContext of the parent #FoFo of the current #FoFo.
 * 
 * Resolve the border properties of @current_context.
 *
 * @parent_context is needed to get the 'color' property value if it
 * is NULL in @current_context.
 **/
void
fo_property_border_resolve (FoContext *current_context,
			    FoContext *parent_context)
{
  FoProperty *absolute, *relative;
  FoDatatype *datatype;
  FoProperty *prop_color;
  FoDatatype *initial_color;

  g_return_if_fail (current_context != NULL);
  g_return_if_fail (parent_context != NULL);
  g_return_if_fail (FO_IS_CONTEXT (current_context));
  g_return_if_fail (FO_IS_CONTEXT (parent_context));

  /* If neither the absolute nor relative color is specified, use the
   * same color as is used for text.
   */
  prop_color = fo_context_get_color (current_context);
  if (prop_color == NULL)
    {
      prop_color = fo_context_get_color (parent_context);
    }

  initial_color = fo_property_get_value (prop_color);

  /* left = start*/
  /* style */
  absolute = fo_context_get_border_left_style (current_context);
  if (absolute != NULL)
    {
      datatype = fo_property_get_value (absolute);
      relative = fo_property_border_start_style_new ();
      fo_property_set_value (relative, datatype);
      fo_context_set_border_start_style (current_context, relative);
    }
  else
    {
      relative = fo_context_get_border_start_style (current_context);
      if (relative != NULL)
	{
	  datatype = fo_property_get_value (relative);
	  absolute = fo_property_border_left_style_new ();
	  fo_property_set_value (absolute, datatype);
	  fo_context_set_border_left_style (current_context, absolute);
	}
      else
	{
	  absolute = fo_property_border_left_style_get_initial ();
	  fo_context_set_border_left_style (current_context, absolute);
	  relative = fo_property_border_start_style_get_initial ();
	  fo_context_set_border_start_style (current_context, relative);
	}
    }
  /* width */
  datatype = fo_property_get_value (absolute);
  if (fo_enum_get_value (datatype) == FO_ENUM_BORDER_STYLE_NONE)
    {
      absolute = g_object_new (FO_TYPE_PROPERTY_BORDER_LEFT_WIDTH,
			       "value",
			       fo_length_get_length_zero (),
			       NULL);
      fo_context_set_border_left_width (current_context, absolute);
      relative = g_object_new (FO_TYPE_PROPERTY_BORDER_START_WIDTH,
			       "value",
			       fo_length_get_length_zero (),
			       NULL);
      fo_context_set_border_start_width (current_context, relative);
    }
  else
    {
      absolute = fo_context_get_border_left_width (current_context);
      if (absolute != NULL)
	{
	  datatype = fo_property_get_value (absolute);
	  relative = fo_property_border_start_width_new ();
	  fo_property_set_value (relative, datatype);
	  fo_context_set_border_start_width (current_context, relative);
	}
      else
	{
	  relative = fo_context_get_border_start_width (current_context);
	  if (relative != NULL)
	    {
	      datatype = fo_property_get_value (relative);
	      absolute = fo_property_border_left_width_new ();
	      fo_property_set_value (absolute, datatype);
	      fo_context_set_border_left_width (current_context, absolute);
	    }
	  else
	    {
	      absolute = fo_property_border_left_width_get_initial ();
	      fo_context_set_border_left_width (current_context, absolute);
	      relative = fo_property_border_start_width_get_initial ();
	      fo_context_set_border_start_width (current_context, relative);
	    }
	}
    }
  /* color */
  absolute = fo_context_get_border_left_color (current_context);
  if (absolute != NULL)
    {
      datatype = fo_property_get_value (absolute);
      relative = fo_property_border_start_color_new ();
      fo_property_set_value (relative, datatype);
      fo_context_set_border_start_color (current_context, relative);
    }
  else
    {
      relative = fo_context_get_border_start_color (current_context);
      if (relative != NULL)
	{
	  datatype = fo_property_get_value (relative);
	  absolute = fo_property_border_left_color_new ();
	  fo_property_set_value (absolute, datatype);
	  fo_context_set_border_left_color (current_context, absolute);
	}
      else
	{
	  prop_color = g_object_new (FO_TYPE_PROPERTY_BORDER_LEFT_COLOR,
				     "value",
				     initial_color,
				     NULL);
	  fo_context_set_border_left_color (current_context, prop_color);
	  prop_color = g_object_new (FO_TYPE_PROPERTY_BORDER_START_COLOR,
				     "value",
				     initial_color,
				     NULL);
	  fo_context_set_border_start_color (current_context, prop_color);
	}
    }
  /* right = end*/
  /* style */
  absolute = fo_context_get_border_right_style (current_context);
  if (absolute != NULL)
    {
      datatype = fo_property_get_value (absolute);
      relative = fo_property_border_end_style_new ();
      fo_property_set_value (relative, datatype);
      fo_context_set_border_end_style (current_context, relative);
    }
  else
    {
      relative = fo_context_get_border_end_style (current_context);
      if (relative != NULL)
	{
	  datatype = fo_property_get_value (relative);
	  absolute = fo_property_border_right_style_new ();
	  fo_property_set_value (absolute, datatype);
	  fo_context_set_border_right_style (current_context, absolute);
	}
      else
	{
	  absolute = fo_property_border_right_style_get_initial ();
	  fo_context_set_border_right_style (current_context, absolute);
	  relative = fo_property_border_end_style_get_initial ();
	  fo_context_set_border_end_style (current_context, relative);
	}
    }
  /* width */
  datatype = fo_property_get_value (absolute);
  if (fo_enum_get_value (datatype) == FO_ENUM_BORDER_STYLE_NONE)
    {
      absolute = g_object_new (FO_TYPE_PROPERTY_BORDER_RIGHT_WIDTH,
			       "value",
			       fo_length_get_length_zero (),
			       NULL);
      fo_context_set_border_right_width (current_context, absolute);
      relative = g_object_new (FO_TYPE_PROPERTY_BORDER_END_WIDTH,
			       "value",
			       fo_length_get_length_zero (),
			       NULL);
      fo_context_set_border_end_width (current_context, relative);
    }
  else
    {
      absolute = fo_context_get_border_right_width (current_context);
      if (absolute != NULL)
	{
	  datatype = fo_property_get_value (absolute);
	  relative = fo_property_border_end_width_new ();
	  fo_property_set_value (relative, datatype);
	  fo_context_set_border_end_width (current_context, relative);
	}
      else
	{
	  relative = fo_context_get_border_end_width (current_context);
	  if (relative != NULL)
	    {
	      datatype = fo_property_get_value (relative);
	      absolute = fo_property_border_right_width_new ();
	      fo_property_set_value (absolute, datatype);
	      fo_context_set_border_right_width (current_context, absolute);
	    }
	  else
	    {
	      absolute = fo_property_border_right_width_get_initial ();
	      fo_context_set_border_right_width (current_context, absolute);
	      relative = fo_property_border_end_width_get_initial ();
	      fo_context_set_border_end_width (current_context, relative);
	    }
	}
    }
  /* color */
  absolute = fo_context_get_border_right_color (current_context);
  if (absolute != NULL)
    {
      datatype = fo_property_get_value (absolute);
      relative = fo_property_border_end_color_new ();
      fo_property_set_value (relative, datatype);
      fo_context_set_border_end_color (current_context, relative);
    }
  else
    {
      relative = fo_context_get_border_end_color (current_context);
      if (relative != NULL)
	{
	  datatype = fo_property_get_value (relative);
	  absolute = fo_property_border_right_color_new ();
	  fo_property_set_value (absolute, datatype);
	  fo_context_set_border_right_color (current_context, absolute);
	}
      else
	{
	  prop_color = g_object_new (FO_TYPE_PROPERTY_BORDER_RIGHT_COLOR,
				     "value",
				     initial_color,
				     NULL);
	  fo_context_set_border_right_color (current_context, prop_color);
	  prop_color = g_object_new (FO_TYPE_PROPERTY_BORDER_END_COLOR,
				     "value",
				     initial_color,
				     NULL);
	  fo_context_set_border_end_color (current_context, prop_color);
	}
    }
  /* top = before*/
  /* style */
  absolute = fo_context_get_border_top_style (current_context);
  if (absolute != NULL)
    {
      datatype = fo_property_get_value (absolute);
      relative = fo_property_border_before_style_new ();
      fo_property_set_value (relative, datatype);
      fo_context_set_border_before_style (current_context, relative);
    }
  else
    {
      relative = fo_context_get_border_before_style (current_context);
      if (relative != NULL)
	{
	  datatype = fo_property_get_value (relative);
	  absolute = fo_property_border_top_style_new ();
	  fo_property_set_value (absolute, datatype);
	  fo_context_set_border_top_style (current_context, absolute);
	}
      else
	{
	  absolute = fo_property_border_top_style_get_initial ();
	  fo_context_set_border_top_style (current_context, absolute);
	  relative = fo_property_border_before_style_get_initial ();
	  fo_context_set_border_before_style (current_context, relative);
	}
    }
  /* width */
  datatype = fo_property_get_value (absolute);
  if (fo_enum_get_value (datatype) == FO_ENUM_BORDER_STYLE_NONE)
    {
      absolute = g_object_new (FO_TYPE_PROPERTY_BORDER_TOP_WIDTH,
			       "value",
			       fo_length_get_length_zero (),
			       NULL);
      fo_context_set_border_top_width (current_context, absolute);
      relative = g_object_new (FO_TYPE_PROPERTY_BORDER_BEFORE_WIDTH,
			       "value",
			       fo_length_get_length_zero (),
			       NULL);
      fo_context_set_border_before_width (current_context, relative);
    }
  else
    {
      absolute = fo_context_get_border_top_width (current_context);
      if (absolute != NULL)
	{
	  datatype = fo_property_get_value (absolute);
	  relative = fo_property_border_before_width_new ();
	  fo_property_set_value (relative, datatype);
	  fo_context_set_border_before_width (current_context, relative);
	}
      else
	{
	  relative = fo_context_get_border_before_width (current_context);
	  if (relative != NULL)
	    {
	      datatype = fo_property_get_value (relative);
	      absolute = fo_property_border_top_width_new ();
	      fo_property_set_value (absolute, datatype);
	      fo_context_set_border_top_width (current_context, absolute);
	    }
	  else
	    {
	      absolute = fo_property_border_top_width_get_initial ();
	      fo_context_set_border_top_width (current_context, absolute);
	      relative = fo_property_border_before_width_get_initial ();
	      fo_context_set_border_before_width (current_context, relative);
	    }
	}
    }
  /* color */
  absolute = fo_context_get_border_top_color (current_context);
  if (absolute != NULL)
    {
      datatype = fo_property_get_value (absolute);
      relative = fo_property_border_before_color_new ();
      fo_property_set_value (relative, datatype);
      fo_context_set_border_before_color (current_context, relative);
    }
  else
    {
      relative = fo_context_get_border_before_color (current_context);
      if (relative != NULL)
	{
	  datatype = fo_property_get_value (relative);
	  absolute = fo_property_border_top_color_new ();
	  fo_property_set_value (absolute, datatype);
	  fo_context_set_border_top_color (current_context, absolute);
	}
      else
	{
	  prop_color = g_object_new (FO_TYPE_PROPERTY_BORDER_TOP_COLOR,
				     "value",
				     initial_color,
				     NULL);
	  fo_context_set_border_top_color (current_context, prop_color);
	  prop_color = g_object_new (FO_TYPE_PROPERTY_BORDER_BEFORE_COLOR,
				     "value",
				     initial_color,
				     NULL);
	  fo_context_set_border_before_color (current_context, prop_color);
	}
    }
  /* bottom = after*/
  /* style */
  absolute = fo_context_get_border_bottom_style (current_context);
  if (absolute != NULL)
    {
      datatype = fo_property_get_value (absolute);
      relative = fo_property_border_after_style_new ();
      fo_property_set_value (relative, datatype);
      fo_context_set_border_after_style (current_context, relative);
    }
  else
    {
      relative = fo_context_get_border_after_style (current_context);
      if (relative != NULL)
	{
	  datatype = fo_property_get_value (relative);
	  absolute = fo_property_border_bottom_style_new ();
	  fo_property_set_value (absolute, datatype);
	  fo_context_set_border_bottom_style (current_context, absolute);
	}
      else
	{
	  absolute = fo_property_border_bottom_style_get_initial ();
	  fo_context_set_border_bottom_style (current_context, absolute);
	  relative = fo_property_border_after_style_get_initial ();
	  fo_context_set_border_after_style (current_context, relative);
	}
    }
  /* width */
  datatype = fo_property_get_value (absolute);
  if (fo_enum_get_value (datatype) == FO_ENUM_BORDER_STYLE_NONE)
    {
      absolute = g_object_new (FO_TYPE_PROPERTY_BORDER_BOTTOM_WIDTH,
			       "value",
			       fo_length_get_length_zero (),
			       NULL);
      fo_context_set_border_bottom_width (current_context, absolute);
      relative = g_object_new (FO_TYPE_PROPERTY_BORDER_AFTER_WIDTH,
			       "value",
			       fo_length_get_length_zero (),
			       NULL);
      fo_context_set_border_after_width (current_context, relative);
    }
  else
    {
      absolute = fo_context_get_border_bottom_width (current_context);
      if (absolute != NULL)
	{
	  datatype = fo_property_get_value (absolute);
	  relative = fo_property_border_after_width_new ();
	  fo_property_set_value (relative, datatype);
	  fo_context_set_border_after_width (current_context, relative);
	}
      else
	{
	  relative = fo_context_get_border_after_width (current_context);
	  if (relative != NULL)
	    {
	      datatype = fo_property_get_value (relative);
	      absolute = fo_property_border_bottom_width_new ();
	      fo_property_set_value (absolute, datatype);
	      fo_context_set_border_bottom_width (current_context, absolute);
	    }
	  else
	    {
	      absolute = fo_property_border_bottom_width_get_initial ();
	      fo_context_set_border_bottom_width (current_context, absolute);
	      relative = fo_property_border_after_width_get_initial ();
	      fo_context_set_border_after_width (current_context, relative);
	    }
	}
    }
  /* color */
  absolute = fo_context_get_border_bottom_color (current_context);
  if (absolute != NULL)
    {
      datatype = fo_property_get_value (absolute);
      relative = fo_property_border_after_color_new ();
      fo_property_set_value (relative, datatype);
      fo_context_set_border_after_color (current_context, relative);
    }
  else
    {
      relative = fo_context_get_border_after_color (current_context);
      if (relative != NULL)
	{
	  datatype = fo_property_get_value (relative);
	  absolute = fo_property_border_bottom_color_new ();
	  fo_property_set_value (absolute, datatype);
	  fo_context_set_border_bottom_color (current_context, absolute);
	}
      else
	{
	  prop_color = g_object_new (FO_TYPE_PROPERTY_BORDER_BOTTOM_COLOR,
				     "value",
				     initial_color,
				     NULL);
	  fo_context_set_border_bottom_color (current_context, prop_color);
	  prop_color = g_object_new (FO_TYPE_PROPERTY_BORDER_AFTER_COLOR,
				     "value",
				     initial_color,
				     NULL);
	  fo_context_set_border_after_color (current_context, prop_color);
	}
    }
}
